import logging
import os

import yaml

from connection import Connection, stats, stats_lock
from bucket import make_bucket
from exception import ConfigParseError, AwsCredentialNotFound

# Silence requests logger
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

log = logging.getLogger('s3tup')

def s3tup(config, access_key_id=None, secret_access_key=None,
          rsync_only=False):
    log.info('**** s3tup ****')

    # Load configuration from filename or file-like object
    try:
        if isinstance(config, basestring):
            config = yaml.load(file(config))
        elif isinstance(config, file):
            config = yaml.load(config)
    except yaml.YAMLError as e:
        raise ConfigParseError(e)

    # Attempt to create Connection from params
    try:
        conn = Connection(access_key_id, secret_access_key)
    except AwsCredentialNotFound:
        conn = None
    for c in config:
        b = make_bucket(conn, **c)
        b.sync(rsync_only)
    
    log.debug('request totals:')
    with stats_lock:
        for k,v in stats.iteritems():
            if v > 0:
                log.debug('- {}: {}'.format(k,v))


class S3Error(Exception):
    pass

class ConfigParseError(Exception):
    pass

class ConfigValidationError(Exception):
    pass

class AwsCredentialNotFound(Exception):
    pass

class SecretAccessKeyNotFound(AwsCredentialNotFound):
    pass

class AccessKeyIdNotFound(AwsCredentialNotFound):
    pass


from base64 import b64encode
from email.utils import formatdate
from threading import Lock
import os
import logging
import hashlib
import hmac
import urllib

from bs4 import BeautifulSoup
from requests import Session, Request
from requests.structures import CaseInsensitiveDict

from exception import S3Error, SecretAccessKeyNotFound, AccessKeyIdNotFound

log = logging.getLogger('s3tup.connection')

stats_lock = Lock()
stats = {'GET':0, 'POST':0, 'PUT':0, 'DELETE':0, 'HEAD':0}

class Connection(object):

    def __init__(self, access_key_id=None, secret_access_key=None):

        if access_key_id is None:
            try: access_key_id = os.environ['AWS_ACCESS_KEY_ID']
            except KeyError:
                raise AccessKeyIdNotFound('Connection object could not be created: You must either supply the access_key_id parameter or set the AWS_ACCESS_KEY_ID environment variable.')

        if secret_access_key is None:
            try: secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']
            except KeyError:
                raise SecretAccessKeyNotFound('Connection object could not be created: You must either supply the secret_access_key parameter or set the AWS_SECRET_ACCESS_KEY environment variable.')

        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key

    def make_request(self, method, bucket, key=None, params=None, data=None,
                     headers=None):

        # Remove params that are set to None
        if isinstance(params, dict):
            for k,v in params.copy().iteritems():
                if v is None:
                    params.pop(k)

        # Construct target url
        url = 'http://{}.s3.amazonaws.com'.format(bucket)
        url += '/{}'.format(key) if key is not None else '/'
        if isinstance(params, dict) and len(params) > 0:
            url += '?{}'.format(urllib.urlencode(params))
        elif isinstance(params, basestring):
            url += '?{}'.format(params)

        # Make headers case insensitive
        if headers is None:
            headers = {}
        headers = CaseInsensitiveDict(headers)

        headers['Host'] = '{}.s3.amazonaws.com'.format(bucket)

        if data is not None:
            m = hashlib.md5()
            m.update(data)
            md5 = b64encode(m.digest())
            headers['Content-MD5'] = md5
        else:
            md5 = ''

        try: content_type = headers['Content-Type']
        except KeyError: content_type = ''

        date = formatdate(timeval=None, localtime=False, usegmt=True)
        headers['x-amz-date'] = date

        # Construct canonicalized amz headers string
        canonicalized_amz_headers = ''
        amz_keys = [k for k in headers.iterkeys() if k.startswith('x-amz-')]
        for k in sorted(amz_keys):
            v = headers[k].strip()
            canonicalized_amz_headers += '{}:{}\n'.format(k.lower(), v)

        # Construct canonicalized resource string
        canonicalized_resource = '/' + bucket
        canonicalized_resource += '/' if key is None else '/{}'.format(key)
        if isinstance(params, basestring):
            canonicalized_resource += '?{}'.format(params)

        # Construct string to sign
        string_to_sign = method.upper() + '\n'
        string_to_sign += md5 + '\n'
        string_to_sign += content_type + '\n'
        string_to_sign += '\n' # date is always set through x-amz-date
        string_to_sign += canonicalized_amz_headers + canonicalized_resource

        # Create signature
        h = hmac.new(self.secret_access_key, string_to_sign, hashlib.sha1)
        signature = b64encode(h.digest())

        # Set authorization header
        headers['Authorization'] = 'AWS {}:{}'.format(self.access_key_id,
                                                      signature)

        # Prepare Request
        s = Session()
        req = Request(method, url, data=data, headers=headers).prepare()

        # Log request data
        log.debug('{} {}'.format(method, url))
        log.debug('headers:')
        for k in sorted(req.headers.iterkeys()):
            log.debug(' {}: {}'.format(k, req.headers[k]))

        # Send request
        resp = s.send(req)
        log.debug('response: {}\n'.format(resp.status_code))

        # Update stats
        with stats_lock:
            stats[method.upper()] += 1
        
        # Handle errors
        if resp.status_code/100 != 2:
            log.error("S3 replied with non 2xx response code!!!!")
            log.error('  request: {} {}'.format(method, url))
            soup = BeautifulSoup(resp.text)
            error = soup.find('error')
            for c in error.children:
                log.error('  {}: {}'.format(c.name, c.text))

            code = error.find('code').text
            message = error.find('message').text
            raise S3Error("{}: {}".format(code, message))

        return resp


import logging
import mimetypes
import os

import utils
import constants

log = logging.getLogger('s3tup.key')

class KeyFactory(object):
    """
    Basically just a container for KeyConfigurators. make_key will create a
    key based on the input parameters and then run each configurator on it
    sequentially, returning a fully configured key ready for sync.

    """

    def __init__(self, configs=[]):
        self.configurators = []
        for c in configs:
            self.add_key_configurator(**c)

    def add_key_configurator(self, **kwargs):
        """Add a configurator to this factory.

        kwargs are everything in constants.KEY_ATTRS plus the matcher fields:
        'patterns', 'ignore_patterns', 'regexes', and 'ignore_regexes'.

        """
        matcher = utils.Matcher(
            kwargs.pop('patterns', None),
            kwargs.pop('ignore_patterns', None),
            kwargs.pop('regexes', None),
            kwargs.pop('ignore_regexes', None),
        )
        self.configurators.append(KeyConfigurator(matcher=matcher, **kwargs))

    def make_key(self, conn, key_name, bucket_name):
        """Return a properly configured key"""
        key = Key(conn, key_name, bucket_name)
        for c in self.configurators:
            if c.effects_key(key.name):
                key = c.configure(key)
        return key


class KeyConfigurator(object):

    def __init__(self, matcher=None, **kwargs):
        self.matcher = matcher

        for k, v in kwargs.iteritems():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                raise TypeError("KeyConfigurator.__init__() got an"
                                " unexpected keyword argument'{}'"
                                 .format(attr))

    def effects_key(self, key_name):
        """Return whether this configurator effects key_name"""
        try: return self.matcher.match(key_name)
        except AttributeError: return True

    def configure(self, key):
        """Return the input key with all configurations applied"""
        for attr in constants.KEY_ATTRS:
            if attr in self.__dict__ and attr != 'metadata':
                key.__dict__[attr] = self.__dict__[attr]

        try: key.metadata.update(self.metadata)
        except AttributeError: pass

        return key


class Key(object):
    """
    Encapsulates configuration for a particular s3 key. It has attributes
    (all defined in constants.KEY_ATTRS) that you can set, delete, modify, 
    and then sync to s3 using the sync or rsync methods.
    """
    
    def __init__(self, conn, name, bucket_name, **kwargs):
        self.conn = conn
        self.name = name
        self.bucket_name = bucket_name

        # Set defaults for required attributes
        self.reduced_redundancy = kwargs.pop('reduced_redundancy', False)
        self.encrypt = kwargs.pop('encrypt', False)
        self.metadata = kwargs.pop('metadata', {})

        for k,v in kwargs.iteritems():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                raise TypeError("Key.__init__() got an unexpected keyword"
                                " argument '{}'".format(attr))

    @property
    def headers(self):
        """Return the headers associated with this key"""
        headers = {}

        try: headers['x-amz-acl'] = self.canned_acl
        except AttributeError: pass

        try:
            if self.reduced_redundancy:
                headers['x-amz-storage-class'] = 'REDUCED_REDUNDANCY'
        except AttributeError: pass

        try: 
            if self.encrypt:
                headers['x-amz-server-side-encryption'] = 'AES256'
        except AttributeError: pass

        for k, v in self.metadata.iteritems():
            headers['x-amz-meta-' + k] = v

        for k in constants.KEY_HEADERS:
            try:
                if self.__dict__[k] is not None:
                    headers[k.replace('_', '-')] = self.__dict__[k]
            except KeyError: pass

        # Guess content-type
        if 'content-type' not in headers:
            content_type_guess = mimetypes.guess_type(self.name)[0]
            if content_type_guess is not None:
                headers['Content-Type'] = content_type_guess

        return headers

    def sync(self):
        """Sync this object's configuration with its respective key on s3"""
        log.info("syncing key '{}'...".format(self.name))

        headers = self.headers
        headers['x-amz-copy-source'] = '/'+self.bucket_name+'/'+self.name
        headers['x-amz-metadata-directive'] = 'REPLACE'

        self.conn.make_request('PUT', self.bucket_name, self.name,
                               headers=headers)
        self.sync_acl()

    def rsync(self, file_like_object):
        """Upload file_like_object to s3 with this object's configuration"""
        log.info("uploading key '{}'...".format(self.name))

        headers = self.headers
        data = file_like_object.read()

        self.conn.make_request('PUT', self.bucket_name, self.name,
                               headers=headers, data=data)
        self.sync_acl()

    def sync_acl(self):
        try: acl = self.acl
        except AttributeError: return False

        if acl is not None:
            self.conn.make_request('PUT', self.bucket_name, self.name, 'acl',
                                   data=acl)
        else:
            self.conn.make_request('PUT', self.bucket_name, self.name, 'acl',
                                   headers={'x-amz-acl': 'private'})
        
        


from fnmatch import fnmatch
from base64 import b64encode
import hashlib
import os
import re

class Matcher(object):
    """Matches patterns"""
    
    def __init__(self, patterns=None, ignore_patterns=None, regexes=None,
                 ignore_regexes=None):
        self.patterns = patterns
        self.ignore_patterns = ignore_patterns
        self.regexes = regexes
        self.ignore_regexes = ignore_regexes

    def match(self, s):
        """Return whether this matcher matches string s"""
        # If neither patterns nor regexes is set, match everything
        matched = self.patterns is None and self.regexes is None 

        if self.patterns is not None:
            for pattern in self.patterns:
                if fnmatch(s, pattern):
                    matched = True
                    break

        if self.regexes is not None and matched is not True:
            for regex in self.regexes:
                if re.search(regex, s):
                    matched = True
                    break

        if self.ignore_patterns is not None and matched is True:
            for pattern in self.ignore_patterns:
                if fnmatch(s, pattern):
                    return False

        if self.ignore_regexes is not None and matched is True:
            for regex in self.ignore_regexes:
                if re.search(regex, s):
                    return False

        return matched


def file_md5(filename):
    """Return md5 hash of file located at filename"""
    m = hashlib.md5()
    with open(filename,'rb') as f:
        while True:
            buf=f.read(8192)
            if not buf: break
            m.update(buf)
    return b64encode(m.digest()).strip()

def os_walk_iter(src):
    """Return list of all file paths in src relative to src"""
    for path, dirs, files in os.walk(src):
        for f in files:
            full_path = os.path.join(path, f)
            yield os.path.relpath(full_path, src)

def key_diff(before, after):
    """
    Return a dict representing the difference between two runs of
    Bucket.get_remote_keys().

    Useful to see changes in the keys within a bucket after some operation.
    Ex:
        before = list(bucket.get_remote_keys())
        do_something()
        diff = key_diff(before, bucket.get_remote_keys())

    Returned dict has fields 'new', 'removed', 'modified', and 'unmodified',
    and each field contains a list of str key names. Size and md5 are used to
    check for modification.

    Note: Be sure that 'before' is a list and *not* the generator returned
    by Bucket.get_remote_keys; because it's a generator, it won't actually
    fetch them from the server until they are iterated through.

    """
    before_set = {k['name'] for k in before}
    after_set = {k['name'] for k in after}

    before_keys = {k['name']: k for k in before if k['name'] in after_set}
    after_keys = {k['name']: k for k in after if k['name'] in before_set}

    removed = before_set - after_set
    new = after_set - before_set

    modified = set()
    unmodified = set()
    for k in (before_set & after_set):
        if before_keys[k]['etag'] == after_keys[k]['etag'] and \
           before_keys[k]['size'] == after_keys[k]['size']:
            unmodified.add(k)
        else:
            modified.add(k)
    return {'new': new, 'removed': removed, 'modified': modified,
            'unmodified': unmodified}



import logging
import os

import utils

log = logging.getLogger('s3tup.rsync')

def rsync(bucket, src='', dest='', delete=False, matcher=utils.Matcher()):
    """Rsync a local path with an s3 bucket.

    Will upload all new and modified local files from the src directory into
    the specified bucket. It uses the key's md5 hash and size to determine
    if it's been modified (and wheter or not to upload), and all keys
    uploaded are properly configured by the inputed bucket.

    Params:

    bucket - s3tup.bucket.Bucket object. The name attribute of this object
             is the destination bucket for this rsync and the keys that
             rsync uses to upload are created and configured using its
             make_key method.

    src - Relative or absolute path to the local folder you want to sync.
          Unlike actual rsyncing, trailing slash does not matter.

    dest - Allows you to rsync to a certain prefix on a bucket.
           Ex: If dest is 'assets', files of the src dir will be rsynced to
           example-bucket.s3.amazonaws.com/assets/

    delete - Whether or not to delete files in your s3 bucket that no longer
             exist locally. When used with dest, only files in dest will be
             deleted.

    matcher - s3tup.utils.Matcher object. If the matcher does not match on
              the local file path, that file will not be rsynced.

    Returns a dict with fields 'new', 'removed', 'modified', and
    'unmodified', and each field contains a list of key names.

    """
    # Silence key logging to avoid redundant messages
    key_log = logging.getLogger('s3tup.key')
    key_log.setLevel(logging.WARNING)

    bucket_url = bucket.name+'.s3.amazonaws.com'
    log.info("rsyncing '{}'".format(os.path.join(bucket_url, dest)))
    log.info("    with '{}'...".format(os.path.abspath(src)))
    
    all_keys = {}

    class RsyncKey(object):

        def __init__(self, name):
            self.name = name
            self.remote = False
            self.local = False

        @property
        def new(self):
            return self.local == True and self.remote == False

        @property
        def removed(self):
            return self.local == False and self.remote == True

        @property
        def modified(self):
            if not (self.local and self.remote):
                return False
            if self.local_md5 != self.remote_md5:
                return True
            if self.local_size != int(self.remote_size):
                return True
            return False

        @property
        def unmodified(self):
            if not (self.local and self.remote):
                return False
            if self.local_md5 != self.remote_md5:
                return False
            if self.local_size != self.remote_size:
                return False
            return True 

        def __lt__(self, other):
            return self.name < other.name

        def rsync(self):
            key = bucket.make_key(self.name)
            key.rsync(open(self.local_path, 'rb'))


    # Local keys
    for relpath in utils.os_walk_iter(src):
        if not matcher.match(relpath):
            continue
        key_name = os.path.join(dest, relpath)
        local_path = os.path.join(src, relpath)

        rsk = RsyncKey(key_name)
        rsk.local = True
        rsk.local_path = local_path
        rsk.local_md5 = utils.file_md5(local_path)
        rsk.local_size = os.path.getsize(local_path)

        all_keys[rsk.name] = rsk

    # Remote keys
    prefix = dest if dest != '' else None
    for key in bucket.get_remote_keys(prefix=prefix):
        try:
            rsk = all_keys[key['name']]
        except KeyError:
            rsk = RsyncKey(key['name'])

        rsk.remote = True
        rsk.remote_md5 = key['etag']
        rsk.remote_size = key['size']

        all_keys[rsk.name] = rsk

    new_keys = []
    removed_keys = []
    modified_keys = []
    unmodified_keys = []

    # Sort every key into appropriate list
    for k in sorted(all_keys.iterkeys()):
        v = all_keys[k]
        if v.new:
            new_keys.append(k)
        elif v.removed:
            removed_keys.append(k)
        elif v.modified:
            modified_keys.append(k)
        elif v.unmodified:
            unmodified_keys.append(k)

    # Delete removed keys if delete
    if delete is True:
        delete_keys(bucket, removed_keys)
    else:
        unmodified_keys.extend(removed_keys)
        removed_keys = []

    # Upload new keys
    for k in new_keys:
        key = all_keys[k]
        log.info("new: '{}', uploading now...".format(k))
        key.rsync()

    # Upload modified keys
    for k in modified_keys:
        key = all_keys[k]
        log.info("modified: '{}', uploading now...".format(k))
        key.rsync()

    key_log.setLevel(logging.DEBUG)

    log.info('rsync complete!')
    log.info('{} new, {} removed, {} modified, {} unmodified'.format(
             len(new_keys), len(removed_keys), len(modified_keys),
             len(unmodified_keys)))

    return {'new': new_keys, 
            'removed': removed_keys, 
            'modified': modified_keys, 
            'unmodified': unmodified_keys}

def delete_keys(bucket, keys):
    """Delete a list of keys from a specified bucket.

    Bucket is a s3tup.bucket.Bucket object and keys is a list of str
    key names to be deleted. S3's delete operation has a limit of
    1000 keys per request so this method handles paging as well.

    """
    for i in xrange(0, len(keys), 1000):
        data = """<?xml version="1.0" encoding="UTF-8"?>
                  <Delete>
                  <Quiet>true</Quiet>\n"""
        for k in keys[i:i+1000]:
            log.info('removed: {}'.format(k))
            data += '<Object><Key>{}</Key></Object>\n'.format(k)
        data += '</Delete>'
        bucket.make_request('POST', 'delete', data=data)


import logging
import binascii

from bs4 import BeautifulSoup

from connection import Connection
from key import KeyFactory, Key
from rsync import rsync
from exception import AwsCredentialNotFound
import utils
import constants

log = logging.getLogger('s3tup.bucket')

def make_bucket(conn=None, **kwargs):
    """Return a properly configured Bucket object from a s3tup configuration.

    Takes every kwarg detailed in the bucket config section of the readme.
    Also takes an optional Connection object as the first parameter. If the
    'access_key_id' and 'secret_access_key' kwargs are not set, this
    connection will be used in the bucket returned. If no connection object
    is passed, it will just call the default Connection constructor which
    will attempt to read your credentials from your env vars.

    """

    bucket_name = kwargs.pop('bucket')

    try:
        access_key_id = kwargs.pop('access_key_id')
        secret_access_key = kwargs.pop('secret_access_key')
        conn = Connection(access_key_id, secret_access_key)
    except KeyError:
        pass

    if conn is None:
        conn = Connection()

    if 'key_config' in kwargs:
        key_factory = KeyFactory(kwargs.pop('key_config'))
    else:
        key_factory = None

    return Bucket(conn, bucket_name, key_factory, **kwargs)


class Bucket(object):
    """
    Encapsulates configuration for an s3 bucket and its keys. It contains a
    key_factory which handles key configuration and many attributes (listed
    in constants.BUCKET_ATTRS) that you can set, delete, modify, and then
    sync to s3 using the various sync methods provided.
    """

    def __init__(self, conn, name, key_factory=None, **kwargs):
        self.conn = conn
        self.name = name
        self.key_factory = key_factory

        self.redirects = kwargs.pop('redirects', [])
        for attr in kwargs:
            if attr in constants.BUCKET_ATTRS:
                self.__dict__[attr] = kwargs[attr]
            else:
                raise TypeError("Bucket.__init__() got an unexpected keyword"
                                " argument '{}'".format(attr))

    def make_key(self, key_name):
        """Return a properly configured Key object."""
        if self.key_factory is None:
            return Key(self.conn, key_name, self.name)
        else:
            return self.key_factory.make_key(self.conn, key_name, self.name)

    def make_request(self, method, params=None, data=None, headers=None):
        """
        Convenience method for self.conn.make_request; has the bucket and
        key fields already filled in.
        """
        return self.conn.make_request(method, self.name, None, params,
                                      data=data, headers=headers)

    # Named get_remote_keys instead of just overriding __iter__
    # to avoid ambiguity. Returns dicts instead of objects for the
    # same reason.
    def get_remote_keys(self, prefix=None):
        """Generate list of dicts representing all keys in this s3 bucket.

        Each dict returned contains fields 'name', 'etag', 'size', and
        'modified'. Paging is handled automatically. Optional (str)
        prefix param will limit the results to those keys prefixed by it.

        """
        more = True
        marker = None
        while more is True:
            params = {'marker': marker, 'prefix': prefix}
            r = self.conn.make_request('GET', self.name, params=params)
            root = BeautifulSoup(r.text).find('listbucketresult')
            for c in root.find_all('contents'):
                key = c.find('key').text
                marker = key
                modified = c.find('lastmodified').text
                size = int(c.find('size').text)
                etag_hex = c.find('etag').text.replace('"', '')
                etag_bin = binascii.unhexlify(etag_hex)
                etag = binascii.b2a_base64(etag_bin).strip()
                yield {'name': key, 'etag': etag, 'size': size,
                       'modified': modified}
            more = root.find('istruncated').text == 'true'

    def sync(self, rsync_only=False):
        """Sync everything.

        Takes every applicable attribute set on this Bucket object and
        configures its respective s3 bucket (defined by self.name) to match
        it. Optional rsync only mode will only run rsync (no setting bucket
        configuration, no syncing unmodified keys, no making redirects).

        """
        log.info("syncing bucket '{}'...".format(self.name))

        # Create the bucket
        try: headers = {'x-amz-acl': self.canned_acl}
        except AttributeError: headers = None
        try:
            data = """<CreateBucketConfiguration 
                   xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
                   <LocationConstraint>{}</LocationConstraint> 
                   </CreateBucketConfiguration >""".format(self.region)
        except AttributeError:
            data = None
        self.conn.make_request('PUT', self.name, headers=headers, data=data)

        if rsync_only and 'rsync' not in self.__dict__:
            log.warning("Running in rsync only mode with no rsync config.")

        if not rsync_only:
            self.sync_bucket()

        # Get the key list before we rsync so we can run diff after and only
        # sync unmodified keys.
        if not rsync_only and self.key_factory is not None:
            before = list(self.get_remote_keys())
        
        self.rsync_keys()

        if not rsync_only and self.key_factory is not None:
            if 'rsync' not in self.__dict__:
                keys = [k['name'] for k in before]
            else:
                after = list(self.get_remote_keys())
                keys = utils.key_diff(before, after)['unmodified']
            self.sync_keys(keys)

        if not rsync_only:
            self.sync_redirects()

        log.info("bucket '{}' sucessfully synced!\n".format(self.name))

    def sync_bucket(self):
        """Run all of the bucket sync methods."""
        self.sync_acl()
        self.sync_cors()
        self.sync_lifecycle()
        self.sync_logging()
        self.sync_notification()
        self.sync_policy()
        self.sync_tagging()
        self.sync_versioning()
        self.sync_website()

    def sync_keys(self, keys=None):
        """Sync every key in keys.

        Keys should be a list of key names. Will not run if self.key_factory
        is None as that implies no key config is set on this object (and it
        would just restore every key to its default). If the keys param is
        not provided, it will run on all keys currently in the s3 bucket. 

        """

        if self.key_factory is None:
            return
        if keys is None:
            keys = [k['name'] for k in self.get_remote_keys()]
        for k in keys:
            key = self.make_key(k)
            key.sync()

    def sync_redirects(self):
        """Create all redirects defined in self.redirects"""
        for k,v in self.redirects:
            log.info("creating redirect from {} to {}".format(k, v))
            headers = {'x-amz-website-redirect-location': v}
            self.conn.make_request('PUT', self.name, k, headers=headers)

    def rsync_keys(self):
        """Run rsync for every rsync config in self.rsync"""
        try: rsync_configs = self.rsync
        except AttributeError: return False

        # Accept either a dict or a list of dicts
        if isinstance(rsync_configs, dict):
            rsync_configs = [rsync_configs,]

        for cfg in rsync_configs:
            matcher = utils.Matcher(
                cfg.pop('patterns', None),
                cfg.pop('ignore_patterns', None),
                cfg.pop('regexes', None),
                cfg.pop('ignore_regexes', None),
            )
            rsync(self, matcher=matcher, **cfg)

    # Individual syncing methods.
    #
    # Each checks if this object has its respective attr set and, if it does,
    # proceeds to sync that value with the s3 bucket. Each will return s3's
    # response in the form of a requests.Response object if the attr is set,
    # and if not will return False. These map directly to the fields
    # defined in the bucket config section of the readme, so if you're
    # looking for details on what values are allowed check there.

    def sync_acl(self):
        try: acl = self.acl
        except AttributeError: return False
        
        if acl is not None:
            log.info("setting bucket acl...")
            return self.make_request('PUT', 'acl', data=acl)
        else:
            log.info("reverting to default bucket acl...")
            return self.make_request('PUT', 'acl',
                                     headers={"x-amz-acl":"private"})

    def sync_cors(self):
        try: cors = self.cors
        except AttributeError: return False

        if cors is not None:
            log.info("setting cors configuration...")
            return self.make_request('PUT', 'cors', data=cors)
        else:
            log.info("deleting cors configuration...")
            return self.make_request('DELETE', 'cors')

    def sync_lifecycle(self):
        try: lifecycle = self.lifecycle
        except AttributeError: return False

        if lifecycle is not None:
            log.info("setting lifecycle configuration...")
            return self.make_request('PUT', 'lifecycle', data=lifecycle)
        else:
            log.info("deleting lifecycle configuration...")
            return self.make_request('DELETE', 'lifecycle')

    def sync_logging(self):
        try: logging = self.logging
        except AttributeError: return False

        if logging is not None:
            log.info("setting logging configuration...")
            data = logging
        else:
            log.info("deleting logging configuration...")
            data = '<?xml version="1.0" encoding="UTF-8"?>\
                    <BucketLoggingStatus \
                    xmlns="http://doc.s3.amazonaws.com/2006-03-01"/>'
        return self.make_request('PUT', 'logging', data=data)

    def sync_notification(self):
        try: notification = self.notification
        except AttributeError: return False

        if notification is not None:
            log.info("setting notification configuration...")
            data = notification
        else:
            log.info("deleting notification configuration...")
            data = '<NotificationConfiguration />'
        return self.make_request('PUT', 'notification', data=data)

    def sync_policy(self):
        try: policy = self.policy
        except AttributeError: return False

        if policy is not None:
            log.info("setting bucket policy...")
            return self.make_request('PUT', 'policy', data=policy)
        else:
            log.info("deleting bucket policy...")
            return self.make_request('DELETE', 'policy')

    def sync_tagging(self):
        try: tagging = self.tagging
        except AttributeError: return False

        if tagging is not None:
            log.info("setting bucket tags...")
            return self.make_request('PUT', 'tagging', data=tagging)
        else:
            log.info("deleting bucket tags...")
            return self.make_request('DELETE', 'tagging')

    def sync_versioning(self):
        try: versioning = self.versioning
        except AttributeError: return False

        if versioning:
            log.info("enabling versioning...")
            status = 'Enabled'
        else:
            log.info("suspending versioning...")
            status = 'Suspended'
        data = '<VersioningConfiguration \
                xmlns="http://s3.amazonaws.com/doc/2006-03-01/">\
                <Status>{}</Status>\
                </VersioningConfiguration>'.format(status)
        return self.make_request('PUT', 'versioning', data=data)

    def sync_website(self):
        try: website = self.website
        except AttributeError: return False

        if website is not None:
            log.info("setting website configuration...")
            return self.make_request('PUT', 'website', data=website)
        else:
            log.info("deleting website configuration...")
            return self.make_request('DELETE', 'website')